////////////////////////////////////////////////////////////////////////////
// NoteCase notes manager project <http://notecase.sf.net>
//
// This code is licensed under BSD license.See "license.txt" for more details.
//
// File: Implements NoteCase HTML format (and NoteCenter format) I/O
////////////////////////////////////////////////////////////////////////////

#include "FormatIOHtml.h"
#include "IOLayerFile64.h"
#include "DocumentIterator.h"
#include "LinkInfo.h"
#include <stdio.h>	//sprintf
#include <stdlib.h>	//sprintf
#include <gtk/gtk.h>	//TOFI glib only?
#ifndef _WIN32
 #include <strings.h>	//strcasecmp
#else
 #ifndef __MINGW32__
  #define snprintf _snprintf
 #endif
 #define strcasecmp stricmp
#endif
#include "debug.h"

int InternalIcon_Name2Index(const char *szName);
const char *InternalIcon_Index2Name(int nIndex);

#define  DOC_PARSE_BLANK		0
#define  DOC_PARSE_EXPECT_TITLE	1
#define  DOC_PARSE_EXPECT_TEXT	2
#define  DOC_PARSE_INSIDE_LINK	3

void replaceall(std::string &strData, const char *szFind, const char *szReplace);
void WriteLevel(DocumentIterator &it, IOProcess &file, int nParentID, bool bHtmlExport = false, bool bCustomCSS = false);

FormatIO_HTML::FormatIO_HTML()
{
	m_nCurParentID = -1;
	m_nCurNodeID = -1;
	m_nDocParseState = DOC_PARSE_BLANK;
	m_bNoteCenterMode = false;
	m_bHtmlExport = false;
	m_nTagsParsedCount = 0;
	m_nLinkStart = -1;
}

FormatIO_HTML::~FormatIO_HTML()
{
}

int FormatIO_HTML::Load(const char *szFile, NoteDocument &doc)
{
	m_objDoc = &doc;  //store pointer

	//clear old document contents
	doc.Clear();

	IOProcess objFile(MODE_READ);
	
	// attach file read layer
	IOLayerFile64 *pFileLayer = new IOLayerFile64;
	objFile.PushLayer(pFileLayer);	

	if(!pFileLayer->m_file.Open(szFile, F64_READ|F64_SHARE_READ|F64_OPEN_EXISTING))
		return DOC_LOAD_NOT_FOUND;

	char szBuffer[1024];
	int nRead;

	//TOFIX better error handling
	bool bContinue = true;
	while(bContinue)
	{
		//NOTE: single layer, no need to Finalize
		// some layers cache data from one step and can return 0 bytes in some cycle
		nRead = objFile.Read(szBuffer, sizeof(szBuffer), bContinue);
		if(nRead > 0)
			Parse(szBuffer, nRead);
	}

	//detect errors in file format (should have at least some HTML tags inside)
	if(0 == m_nTagsParsedCount)
		return DOC_LOAD_FORMAT_ERROR;

	return DOC_LOAD_OK;
}

bool FormatIO_HTML::Save(const char *szFile, NoteDocument &doc)
{
	m_objDoc = &doc;  //store pointer

	IOProcess objFile(MODE_WRITE);

	// attach file write layer
	IOLayerFile64 *pFileLayer = new IOLayerFile64;
	objFile.PushLayer(pFileLayer);	

	if(!pFileLayer->m_file.Open(szFile, F64_WRITE|F64_SHARE_READ|F64_OPEN_NEW))
		return false;

	WriteDocContents(objFile);

	return true;
}

void FormatIO_HTML::WriteDocContents(IOProcess &objFile)
{
	ASSERT(NULL != m_objDoc);

	//write header
	if(!m_bHtmlExport)
	{
		if(m_bNoteCenterMode)
			objFile.WriteString("<!DOCTYPE NoteCenter-File>\r\n");
		else
			objFile.WriteString("<!DOCTYPE NoteCase-File>\r\n");
	
		//calc and write selected node "index"
		char szLastNote[100];
		snprintf(szLastNote, sizeof(szLastNote), "<!--LastNote:%d-->\r\n", m_objDoc->m_nCurrentNode);
		objFile.WriteString(szLastNote);
	}

	//write inital HTML tags including HEAD section
	objFile.WriteString("<HTML>\r\n<HEAD>\r\n");
	//define used charset
	objFile.WriteString("<meta content=\"text/html;charset=UTF-8\" http-equiv=\"Content-Type\">\r\n");
	objFile.WriteString("<TITLE>");
	//TOFIX write file title as document title???
	objFile.WriteString("</TITLE>\r\n");

	if(m_bHtmlExport)
	{
		if(m_objDoc->m_strCustomCSS.size() > 0)
		{
			objFile.WriteString("<link rel=\"stylesheet\" type=\"text/css\" href=\"");
			objFile.WriteString(m_objDoc->m_strCustomCSS.c_str());
			objFile.WriteString("\">\r\n");
		}
	}

	objFile.WriteString("</HEAD>\r\n<BODY>\r\n");

	//recursively iterate through document node tree
	DocumentIterator it(*m_objDoc);
	WriteLevel(it, objFile, -1, m_bHtmlExport, m_objDoc->m_strCustomCSS.size() > 0);

	//write footer
	objFile.WriteString("</BODY>\r\n</HTML>\r\n");
}

// HTMLParser event handlers
void FormatIO_HTML::OnTagBegin(const char *szTag, const char *szParams)
{
	//TRACE("FormatIO_HTML:[%s] tag start\n", szTag);

	m_nTagsParsedCount ++;

	//if proper tag, create new node
	if(0 == strcasecmp(szTag, "DL"))	//DL indicates sublevel branching
	{
		m_nCurParentID = m_nCurNodeID;
		m_nCurNodeID   = -1;	
	}
	else if(0 == strcasecmp(szTag, "DT"))	//DT indicates new node title coming
	{
		m_objDoc->NodeInsert(m_nCurParentID, -1);  //TOFIX error report
		m_nCurNodeID = m_objDoc->GetNodeByIdx(m_objDoc->GetNodeCount()-1).m_nID;
		m_nDocParseState = DOC_PARSE_EXPECT_TITLE;
	}
	else if(0 == strcasecmp(szTag, "DD"))	//DD indicates new node text coming
	{
		m_nDocParseState = DOC_PARSE_EXPECT_TEXT;
	}
	else if(0 == strcasecmp(szTag, "A"))	//A indicates hyperlink
	{
		m_nDocParseState = DOC_PARSE_INSIDE_LINK;
		m_strLinkURL = szParams;
		std::string::size_type nPos = m_strLinkURL.find("href=\"");
		if(nPos != std::string::npos)
			m_strLinkURL = m_strLinkURL.substr(nPos+strlen("href=\""));
		if(m_strLinkURL.at(m_strLinkURL.size()-1) == '\"')
			m_strLinkURL = m_strLinkURL.substr(0, m_strLinkURL.size()-1);

		m_nLinkStart = -1;
		int nIdx = m_objDoc->GetIdxFromID(m_nCurNodeID);
		if(nIdx >= 0)
			m_nLinkStart = g_utf8_strlen(m_objDoc->GetNodeByIdx(nIdx).GetText().c_str(), -1);
	}
	else if(0 == strcasecmp(szTag, "BR")) // line break
	{
		int nIdx = m_objDoc->GetIdxFromID(m_nCurNodeID);
		if(nIdx >= 0)
		{
			std::string strTxt = m_objDoc->GetNodeByIdx(nIdx).GetText(); //TOFIX why not return by adress!?
			strTxt += "\r\n";
			m_objDoc->GetNodeByIdx(nIdx).SetText(strTxt.c_str());
		}	
	}
}

void FormatIO_HTML::OnTagEnd(const char *szTag)	
{
	//TRACE("FormatIO_HTML:[%s] tag end\n", szTag);

	//if proper tag, position on the parent node ?
	if(0 == strcasecmp(szTag, "DL"))	// </DL> indicates end of sublevel branching
	{
		int nIdx = m_objDoc->GetIdxFromID(m_nCurParentID);
		if(nIdx >= 0)
			m_nCurParentID = m_objDoc->GetNodeByIdx(nIdx).m_nParentID;
		//m_nCurNodeID   = -1;
	}
	else if(0 == strcasecmp(szTag, "DT"))	//DT indicates new node title coming
	{
	}
	else if(0 == strcasecmp(szTag, "A"))	//A indicates hyperlink
	{
		int nIdx = m_objDoc->GetIdxFromID(m_nCurNodeID);
		if(nIdx >= 0)
		{
			LinkInfo info;
			info.m_strTargetURL = m_strLinkURL;
			info.m_nStartOffset = m_nLinkStart;
			info.m_strText      = m_strLinkText;
			info.RefreshLength();

			m_objDoc->GetNodeByIdx(nIdx).m_lstLinks.push_back(info);

			//clear internal data
			m_strLinkURL = "";
			m_strLinkText = "";
			m_nLinkStart = -1;
		}

		m_nDocParseState = DOC_PARSE_EXPECT_TEXT; // more text?
	}
}

void FormatIO_HTML::OnComment(const char *szText)
{
	//TRACE("FormatIO_HTML:[%s] HTML comment\n", szText);

	//extract info from comments <!--LastNote:4-->
	if(0 == strncmp(szText, "LastNote:", strlen("LastNote:")))
	{
		m_objDoc->m_nCurrentNode = atoi(szText+strlen("LastNote:"));
	}
	else if (0 == strncmp(szText, "property:icon_file=", strlen("property:icon_file=")))
	{
		//example tag: "<!--property:icon_file=/root/aaa.xpm-->"
		//TRACE("icon_file->%s\n",szText);
		DocumentIterator it(*m_objDoc);
		it.GetNodeByID(m_nCurNodeID).m_nIconType = ICON_CUSTOM;
		it.GetNodeByID(m_nCurNodeID).m_strIconFile = szText + strlen("property:icon_file=");
	}
	else if (0 == strncmp(szText, "property:icon_internal=", strlen("property:icon_internal=")))
	{
		//TRACE("icon_internal->%s\n",szText);
		//example tag: "<!--property:icon_internal=blank-->"
		DocumentIterator it(*m_objDoc);
		int nIdx = InternalIcon_Name2Index(szText + strlen("property:icon_internal="));		
		if(nIdx >= 0)
			it.GetNodeByID(m_nCurNodeID).m_nIconType = nIdx;
	}
	else if (0 == strncmp(szText, "property:keywords=", strlen("property:keywords=")))
	{
		//append keywords (multiple tags can exist)
		std::string strKeywords;

		DocumentIterator it(*m_objDoc);
		strKeywords  = it.GetNodeByID(m_nCurNodeID).GetKeywords();
		strKeywords += ";";
		strKeywords += szText + strlen("property:keywords=");

		it.GetNodeByID(m_nCurNodeID).SetKeywords(strKeywords.c_str());
	}
	else if (0 == strncmp(szText, "property:date_created=", strlen("property:date_created=")))
	{
		std::string strDate;
		strDate = szText + strlen("property:date_created=");

		DocumentIterator it(*m_objDoc);
		it.GetNodeByID(m_nCurNodeID).m_nDateCreated = atol(strDate.c_str());
	}
	else if (0 == strncmp(szText, "property:date_modified=", strlen("property:date_modified=")))
	{
		std::string strDate;
		strDate = szText + strlen("property:date_modified=");

		DocumentIterator it(*m_objDoc);
		it.GetNodeByID(m_nCurNodeID).m_nDateModified = atol(strDate.c_str());
	}
	else if (0 == strncmp(szText, "property:finished=", strlen("property:finished=")))
	{
		std::string strDate;
		strDate = szText + strlen("property:finished=");

		DocumentIterator it(*m_objDoc);
		it.GetNodeByID(m_nCurNodeID).m_bFinished = (atoi(strDate.c_str()) != 0);
	}
}

void FormatIO_HTML::OnText(const char *szText)
{
	std::string data(szText);

	//TRACE("FormatIO_HTML:[%s] HTML text\n", szText);

	//set text or title for current node
	if(DOC_PARSE_EXPECT_TITLE == m_nDocParseState)
	{
		//TRACE("FormatIO_HTML:[%s] is node title \n", szText);

		if(!m_bNoteCenterMode)
			HTMLParser::UnescapeChars(data);

		//set node title
		int nIdx = m_objDoc->GetIdxFromID(m_nCurNodeID);
		if(nIdx >= 0)
			m_objDoc->GetNodeByIdx(nIdx).SetTitle(data.c_str());

		m_nDocParseState = DOC_PARSE_BLANK;
	}
	else if(DOC_PARSE_EXPECT_TEXT == m_nDocParseState)
	{
		//TRACE("FormatIO_HTML:[%s] is node text\n", szText);

		HTMLParser::UnescapeChars(data);	//only for node text (NC format)

		//append node text
		int nIdx = m_objDoc->GetIdxFromID(m_nCurNodeID);
		if(nIdx >= 0){
			//NoteCenter Text format: remove \r\n from text start and text end
			if(0 == strncmp(data.c_str(), "\r\n", 2))
				data = data.substr(2, data.size()-2);
			if(0 == strncmp(data.c_str()+data.size()-2, "\r\n", 2))
				data = data.substr(0, data.size()-2);

			//replaceall(strData, "\r\n", "<br>\r\n");

			std::string strCurText = m_objDoc->GetNodeByIdx(nIdx).GetText();
			strCurText += data;

			m_objDoc->GetNodeByIdx(nIdx).SetText(strCurText.c_str());
		}

		//m_nDocParseState = DOC_PARSE_BLANK;
	}
	else if(DOC_PARSE_INSIDE_LINK == m_nDocParseState)
	{
		HTMLParser::UnescapeChars(data);	//only for node text (NC format)
		m_strLinkText = data;

		//append node text
		int nIdx = m_objDoc->GetIdxFromID(m_nCurNodeID);
		if(nIdx >= 0){
			std::string strCurText = m_objDoc->GetNodeByIdx(nIdx).GetText();
			strCurText += data;
			m_objDoc->GetNodeByIdx(nIdx).SetText(strCurText.c_str());
		}
	}
}

void WriteLevel(DocumentIterator &it, IOProcess &file, int nParentID, bool bHtmlExport, bool bCustomCSS)
{
	std::string strData;
	int nSiblingIdx = 0;

	int nIdx = it.GetChildIdx(nParentID, 0);  //first child
	if(nIdx < 0)
		return;

	file.WriteString("<DL>\r\n");

	while(nIdx >= 0)
	{
		NoteNode &node = it.GetNodeByIdx(nIdx);

		//write node title
		strData = node.GetTitle();
		//TOFIX if(!m_bNoteCenterMode)
			HTMLParser::EscapeChars(strData);

		if(bHtmlExport && !bCustomCSS)
			file.WriteString("<DT style=\"font-weight: bold;\">");
		else
			file.WriteString("<DT>");
		file.WriteString(strData.c_str());
		file.WriteString("</DT>\r\n");

		//write node parameters
		if(!bHtmlExport)
		{
			//write node icon setting
			if(ICON_NONE != node.m_nIconType)
			{
				if(ICON_CUSTOM == node.m_nIconType)
				{
					file.WriteString("<!--property:icon_file=");
					file.WriteString(node.m_strIconFile.c_str());
					file.WriteString("-->\r\n");
				}
				else
				{
					file.WriteString("<!--property:icon_internal=");
					file.WriteString(InternalIcon_Index2Name(node.m_nIconType));
					file.WriteString("-->\r\n");
				}
			}

			//write node keywords
			std::string strKeywords = node.GetKeywords();
			if(strKeywords.size()>0)
			{
					file.WriteString("<!--property:keywords=");
					file.WriteString(strKeywords.c_str());
					file.WriteString("-->\r\n");
			}

			//write node dates
			char szBuffer[20];
			snprintf(szBuffer, sizeof(szBuffer), "%d", (int)node.m_nDateCreated);//TOFIX? (long)
			file.WriteString("<!--property:date_created=");
			file.WriteString(szBuffer);
			file.WriteString("-->\r\n");

			snprintf(szBuffer, sizeof(szBuffer), "%d", (int)node.m_nDateModified); //TOFIX? (long)
			file.WriteString("<!--property:date_modified=");
			file.WriteString(szBuffer);
			file.WriteString("-->\r\n");
		
			snprintf(szBuffer, sizeof(szBuffer), "%d", node.m_bFinished);
			file.WriteString("<!--property:finished=");
			file.WriteString(szBuffer);
			file.WriteString("-->\r\n");
		}

		//write node text
		strData = node.GetText();

		//before escaping, modify contents to include all node links 
		//<A href="http://belle.monitor.hr/0510.htm">Belle de Jour</A>
		int nLinks = node.m_lstLinks.size();
		if(nLinks > 0)
		{
			//sort links by offset in ascending order
			node.m_lstLinks.Sort();

			//break node text into segments and then insert the links tags to the right places
			std::vector<std::string> lstSegments;
			int i;
			for(i=0; i<nLinks; i++)
			{
				LinkInfo &link = node.m_lstLinks[i];

				if(0 == i)
				{
					//convert character offset to the buffer offset
					int nPosBuf;
					const char *szBuf = strData.c_str();
				#ifdef _DEBUG
					int nCharLen      = g_utf8_strlen(szBuf, -1);
					ASSERT(nCharLen > link.m_nStartOffset);
				#endif
					gchar* pszPos = g_utf8_offset_to_pointer(szBuf, link.m_nStartOffset);
					nPosBuf = pszPos - szBuf;

					if(nPosBuf > 0){
						std::string strStart = strData.substr(0, nPosBuf);
						lstSegments.push_back(strStart);
					}
				
					std::string strLink;
					strLink += "<A href=\"";
					strLink += link.m_strTargetURL;
					strLink += "\">";
					lstSegments.push_back(strLink);

					strLink = link.m_strText;
					lstSegments.push_back(strLink);

					strLink = "</A>";
					lstSegments.push_back(strLink);

					std::string strEnd = strData.substr(nPosBuf + link.m_strText.size());
					if(strEnd.size() > 0)
						lstSegments.push_back(strEnd);
				}
				else
				{
					//
					//keep breaking within the segments list
					//

					//find segment for given offset (ignore link segments)
					int nSegmentIdx = -1;
					int nOffset = link.m_nStartOffset;		
					int nOffsetTot = 0;
					for(unsigned int j=0; j<lstSegments.size(); j++)
					{
						std::string &strTmp = lstSegments[j];
						if(strTmp.at(0) == '<')	//skip link tags
							continue;

						int nSize = g_utf8_strlen(strTmp.c_str(), -1);
						if(nOffsetTot + nSize > nOffset){
							nSegmentIdx = j;
							break;
						}
						nOffsetTot += nSize; //keep looking
					}

					//now break the segment into subsegments
					if(nSegmentIdx >= 0)
					{
						std::string strTmp = lstSegments[nSegmentIdx];

						//convert character offset to the buffer offset
						int nPosBuf;
						const char *szBuf = strTmp.c_str();
						gchar* pszPos = g_utf8_offset_to_pointer(szBuf, link.m_nStartOffset-nOffsetTot);
						nPosBuf = pszPos - szBuf;
	
						//erase original segment
						lstSegments.erase(lstSegments.begin()+nSegmentIdx);

						//replace it with up to 3 new segments (content before,link,content after)
						if(nPosBuf>0){
							lstSegments.insert(lstSegments.begin()+nSegmentIdx, strTmp.substr(0, nPosBuf));
							nSegmentIdx++;
						}
						//link segment
						std::string strLink;
						strLink += "<A href=\"";
						strLink += link.m_strTargetURL;
						strLink += "\">";
						lstSegments.insert(lstSegments.begin()+nSegmentIdx, strLink);
						nSegmentIdx++;

						strLink = link.m_strText;
						lstSegments.insert(lstSegments.begin()+nSegmentIdx, strLink);
						nSegmentIdx++;

						strLink = "</A>";
						lstSegments.insert(lstSegments.begin()+nSegmentIdx, strLink);
						nSegmentIdx++;

						//reamaining data
						strTmp = strTmp.substr(nPosBuf+link.m_strText.size());
						if(strTmp.size() > 0)
							lstSegments.insert(lstSegments.begin()+nSegmentIdx, strTmp);
					}
				}
			}

			//escape segments (skipping link tags) and concatenate back to single string
			strData = "";
			for(unsigned int k=0; k<lstSegments.size(); k++)
			{
				std::string &strTmp = lstSegments[k];
				ASSERT(strTmp.size() > 0);
				if(strTmp.at(0) != '<')	//skip link tags
					HTMLParser::EscapeChars(strTmp);
				strData += strTmp;
			}
		}
		else
		{
			HTMLParser::EscapeChars(strData);
		}

		//change all occurences "\r" to "\r\n"
		replaceall(strData, "\r\n", "\r");
		replaceall(strData, "\n", "\r");
		replaceall(strData, "\r", "\r\n");

		//add tag for line ends
		replaceall(strData, "\r\n", "<BR>\r\n"); //TOFIX

		file.WriteString("<DD>\r\n");
		file.WriteString(strData.c_str());

		//recursively write children of this node
		WriteLevel(it, file, node.m_nID, bHtmlExport, bCustomCSS);

		file.WriteString("\r\n</DD>\r\n");

		//go to the next sibling node
		nSiblingIdx ++;
		nIdx = it.GetChildIdx(nParentID, nSiblingIdx);
	}

	//
	file.WriteString("</DL>\r\n");
}

void replaceall(std::string &strData, const char *szFind, const char *szReplace)
{
//	strData.replace(strData.begin(), strData.end(), szFind, szReplace);

	int nSrcSize = strlen(szFind);
	if(nSrcSize > 0)
	{
		int nDstSize = strlen(szReplace);

		size_t nPos = strData.find(szFind);
		while(nPos != std::string::npos)
		{
			//strData.replace(nPos, nSrcSize, szReplace);
			strData.erase(nPos, nSrcSize);
			strData.insert(nPos, szReplace);

			nPos = strData.find(szFind, nPos+nDstSize); //keep looking forward
		}
	}
}
