////////////////////////////////////////////////////////////////////////////
// NoteCase notes manager project <http://notecase.sf.net>
//
// This code is licensed under BSD license.See "license.txt" for more details.
//
// File: Main method implementation (application startup)
////////////////////////////////////////////////////////////////////////////

#include "config.h"
#include "mru.h"
#include <gtk/gtk.h>
#include <libintl.h>

#include "interface.h"
#include "callbacks.h"
#include "support.h"
#include "lib/IniFile.h"
#include "lib/FilePath.h"
#include "lib/NoteDocument.h"
#include "gui/GuiLanguage.h"
#include "MainWnd.h"

#ifdef _WIN32
 #include <io.h> //access
 #define access _access
#endif

#ifdef _WIN32
 #include "_win/SingleInstance.h"
 CSingleInstance g_instance("NoteCase");
#else
 #include "_unx/SingleInstance.h"
 CSingleInstance g_instance("/NoteCase");
#endif

GtkWidget *window1;
extern MRU g_objMRU;
bool g_bMinimizeToTray = false;
extern MainWindow g_wnd;

int  load_file(const char *filename, const char *szPassword = NULL);
void export_document(int nExportMode, const std::string &strExportFile, const std::string &strCSS, bool bCheckOverwrite = true);

#ifdef HAVE_GNOME_VFS
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-mime-handlers.h>
#endif

const char *GetIniFile()
{
	static std::string strDir;
	strDir  = GetHomeDir();
	strDir += "/.notecase/notecase.ini";
	return strDir.c_str();
}

std::string GetHelpFile()
{
	//prepare help document path
	std::string strHelp;

#ifndef _WIN32
	strHelp = INSTALL_PREFIX;
	strHelp += HELP_FILE_PATH;
#endif

	//if file not defined in a fixed location or not found calc path dynamically
	if( strHelp.empty() ||
	    0 != access(strHelp.c_str(), 0)) 
	{
		//Win32 and Linux debug version
		strHelp = GetAppPath();
		strHelp = GetParentDir(strHelp.c_str());
		strHelp += "help.ncd";
	}
	
	return strHelp;
}

const char *calculate_autosave_filename()
{
	//there exists one autosave file per user
	static std::string strDir;
	strDir  = GetHomeDir();
#ifdef _WIN32
	strDir += "\\.notecase\\autosave.ncd";
#else
	strDir += "/.notecase/autosave.ncd";
#endif

	return strDir.c_str();
}

//this is a version for encrypted format
const char *calculate_autosave_filename1()
{
	//there exists one autosave file per user
	static std::string strDir;
	strDir  = GetHomeDir();
#ifdef _WIN32
	strDir += "\\.notecase\\autosave.nce";
#else
	strDir += "/.notecase/autosave.nce";
#endif

	return strDir.c_str();
}

const char *get_locale_dir()
{
	static std::string strDir;
#ifdef _WIN32
	strDir  = GetAppPath();
	strDir  = GetParentDir(strDir.c_str());
	strDir += "locale/";
#else
	strDir  = INSTALL_PREFIX;
	strDir += LOCALE_DIR;
#endif
	return strDir.c_str();
}

GuiLanguage g_lang(PACKAGE, get_locale_dir());

int main (int argc, char *argv[])
{
	//check startup options
	IniFile file;
	file.Load(GetIniFile());

	//initialize language system
	std::string strLocale;
	file.GetValue("Display", "Language", strLocale, "");
	g_lang.Initialize(strLocale.c_str());

	gtk_init (&argc, &argv);
	
	bool bAllowSingleInst = false;
	file.GetValue("Startup", "AllowSingleInstance", bAllowSingleInst);
	if(bAllowSingleInst && g_instance.ProgramAlreadyStarted())
	{
		gtkMessageBox(_("Another NoteCase instance detected! Exiting!"));
		return 0;
	}

	g_wnd.Create();

	bool bBackupLoaded = autosave_check_crash();

	// check for command line arguments: file name to open
	if(!bBackupLoaded)
	{
		bool bExportFile = false;
		std::string strExportFile;
		std::string strExportCSS;
		std::string strPassword;

		if(argc > 1)
		{
			for(int i=1; i<argc; i++)
			{
				if(0 == strcmp(argv[i], "--help"))
				{
					on_help1_activate (NULL, NULL);	//load help
				}
				else if(0 == strcmp(argv[i], "--password"))
				{
					if(i+1 < argc){ //next token must be password
						strPassword = argv[i+1];
						i++; continue;
					}
					else break;	//invalid flag usage
				}
				else if(0 == strcmp(argv[i], "--export-file"))
				{
					if(i+1 < argc){ //next token must be file name
						strExportFile = argv[i+1];
						bExportFile = true;
						i++; continue;
					}
					else break;	//invalid flag usage
				}
				else if(0 == strcmp(argv[i], "--export-css"))
				{
					if(i+1 < argc){ //next token must be CSS path
						strExportCSS = argv[i+1];
						bExportFile = true;
						i++; continue;
					}
					else break;	//invalid flag usage
				}
				else
				{
					int nResult = load_file(argv[i], strPassword.c_str());
					g_objMRU.Change(argv[1], (DOC_LOAD_NOT_FOUND != nResult));//update MRU unless file not found
				}
			}

			if(bExportFile)
			{
				export_document(2 /*entire doc*/, strExportFile, strExportCSS, false);
				return 0;
			}
		}
		else{
			//check "reload last file" option
			bool bLoadLast;
			if(file.GetValue("Startup", "ReloadLastDocument", bLoadLast))
				if(bLoadLast && (g_objMRU.GetSize() > 0))
					load_file(g_objMRU.GetFile(0));
		}
	}

#ifdef HAVE_GNOME_VFS
	gnome_vfs_init();
#endif

	gtk_main ();

#ifdef HAVE_GNOME_VFS
	gnome_vfs_shutdown();
#endif
	
	//TOFIX move this to quit handler
	autosave_shutdown();
	g_objMRU.Save();
	return 0;
}

#ifdef _WIN32
#include <windows.h>
int APIENTRY WinMain( HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR lpCmdLine, int nCmdShow)
{
	return main( __argc, __argv );
}
#endif

